// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include <stdexcept>
#include <iostream>
#include <sstream>
#include <string>
#include <cstring>
#include <tetgen_mesh.h>

#include "plc.h"

using namespace tritetmesh;

//-----------------------------------------------------------------------------
PLC::PLC():
  radius_edge_ratio(2.0),
  uniform_volume_constraint(-1.0),
  verbosity(1),   
  split_facets(true),
  use_regional_attributes(false),
  use_regional_volume_constraints(false),
  data()
{}
//-----------------------------------------------------------------------------
PLC::~PLC()
{
}
//-----------------------------------------------------------------------------
void PLC::test(bool save_test_result)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: PLC empty. Nothing to test." );
  char tmp0[] = "d";
  char tmp2[12];
  strcpy(tmp2,tmp0);
  tetgen::tetgenio dummy;
  tetgen::tetrahedralize(tmp2, &data, &dummy);
  if (save_test_result)
    {
      char tmp1[] = "test_result";
      strcpy(tmp2,tmp1);
      dummy.save_nodes(tmp2);
      dummy.save_faces(tmp2);
    }
}
//-----------------------------------------------------------------------------
void PLC::load(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  clear();
  enum objecttype {NONE, NODES, POLY, OFF, PLY, STL, MEDIT, MESH};
  enum objecttype object;
  object = NONE;
  // Recognize the object from file extension if it is available.
  // Note: (Code ripped directly from tetgen.cxx)
  if (!strcmp(&file_name[strlen(file_name) - 5], ".poly")) {
    file_name[strlen(file_name) - 5] = '\0';
    object = POLY;
  } else if (!strcmp(&file_name[strlen(file_name) - 6], ".smesh")) {
    file_name[strlen(file_name) - 6] = '\0';
    object = POLY;
  } else if (!strcmp(&file_name[strlen(file_name) - 4], ".off")) {
    file_name[strlen(file_name) - 4] = '\0';
    object = OFF;
  } else if (!strcmp(&file_name[strlen(file_name) - 4], ".ply")) {
    file_name[strlen(file_name) - 4] = '\0';
    object = PLY;
  } else if (!strcmp(&file_name[strlen(file_name) - 4], ".stl")) {
    file_name[strlen(file_name) - 4] = '\0';
    object = STL;
  } else if (!strcmp(&file_name[strlen(file_name) - 5], ".mesh")) {
    file_name[strlen(file_name) - 5] = '\0';
    object = MEDIT;
  }
  data.load_plc(file_name, object);
}
//-----------------------------------------------------------------------------
void PLC::save_tetgen(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: PLC empty. Nothing to save." );
  data.save_nodes(file_name);
  data.save_poly(file_name);
}
//-----------------------------------------------------------------------------
void PLC::clear()
{
  data.deinitialize();
  data.initialize();
}
//-----------------------------------------------------------------------------
const PLC& PLC::operator=(const PLC& plc)
{
  data = plc.data;
  return *this;
}
//-----------------------------------------------------------------------------
void PLC::print_tetgen_parameters()
{
  std::cout << std::endl;
  std::cout << "Tetgen parameters: " << std::endl;
  std::cout << "+----------------------------------------------" << std::endl;     
  std::cout << "| Radius to edge ratio            (q) : " << radius_edge_ratio << std::endl;
  std::cout << "| Uniform volume constraint    (aX.X) : " << uniform_volume_constraint << std::endl;
  std::cout << "| Verbosity                       (V) : " << verbosity << std::endl;
  std::cout << "| Split facets                    (Y) : " << (split_facets ? "True": "False") << std::endl;
  std::cout << "| Use regional attributes         (A) : " << (use_regional_attributes  ? "True": "False") << std::endl;
  std::cout << "| Use regional volume constraints (a) : " << (use_regional_volume_constraints  ? "True": "False") << std::endl;
  std::cout << "+----------------------------------------------" << std::endl << std::endl;     
}
//-----------------------------------------------------------------------------
// FIXME: Remove when proved not needed
void PLC::build_mesh(TetMesh &output, std::string in_switches)
{
  char switches[1024];
  strcpy(switches, in_switches.c_str());

  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: PLC empty. Nothing to build." );
  output.clear();
  
  tetgen::tetrahedralize(switches, &data, &output.data);
}
//-----------------------------------------------------------------------------
void PLC::build_mesh(TetMesh &output)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: PLC empty. Nothing to build." );
  output.clear();
  
  if (verbosity > 0)
    {
      print_tetgen_parameters();
    }
  
  /// Build the switch string
  std::string switches("pz");

  if (use_regional_attributes)
    {
      switches += "A";
    }	   

  if (use_regional_volume_constraints)
    {
      switches += "a";
    }
  
  if (radius_edge_ratio > 1.0) 
    {
      switches += "q";
      std::ostringstream o;
      if (!(o << radius_edge_ratio))
	throw std::runtime_error ("Couldn't convert double to string");
      switches += o.str();
    }
  
  if (uniform_volume_constraint > 0.0)
    {
      switches += "a";
      std::ostringstream o;
      if (!(o << uniform_volume_constraint))
	throw std::runtime_error ("Couldn't convert double to string");
      switches += o.str();
    }
  
  if (!split_facets)
    {
      switches += "Y";
    }

  if ( 0 <= verbosity && verbosity <= 3)
    {
      if (!(verbosity == 1))
	{
	  if (verbosity == 0) 
	    {
	      switches += "Q";
	    }
	  else 
	    for (int i = 1; i <= verbosity; i++ )
	      switches += "V";
	}
    }
  
  // Convert the const char* to a char* 
  char cstr_switches[switches.size()];
  std::strcpy(cstr_switches,switches.c_str());
  std::cout << "Tetgen switches: " << cstr_switches << std::endl;
  
  tetgen::tetrahedralize(cstr_switches, &data, &output.data);
}
//-----------------------------------------------------------------------------



