// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include <cstring>
#include <cstdlib>
#include <stdio.h>
#include <iostream>
#include <algorithm>
#include <stdexcept>
#include <tetgen_mesh.h>

#include "plceditor.h"
#include "plc.h"

using namespace tritetmesh;

//-----------------------------------------------------------------------------
PLCEditor::PLCEditor():
  plc_builded(false),
  tetmesh_builded(false)
{}
//-----------------------------------------------------------------------------
PLCEditor::~PLCEditor()
{}
//-----------------------------------------------------------------------------
void PLCEditor::print_vertices()
{
  if (x_coord.size() == 0)
  {
    std::cout << "*** Warning: No vertices registrated." << std::endl;
  }
  for ( unsigned int i = 0; i < x_coord.size(); ++i)
  {
    std::cout << "Vertex " << i << ": (" << x_coord[i] << ", " 
	      << y_coord[i] << ", " << z_coord[i] << ")"<< std::endl;
  }
}
//-----------------------------------------------------------------------------
void PLCEditor::register_vertices(const std::vector<double> &x_coord,
				  const std::vector<double> &y_coord,
				  const std::vector<double> &z_coord)
{
  if ( ! ( x_coord.size() == y_coord.size() ) ||
       ! ( x_coord.size() == z_coord.size() ) || 
       ! ( y_coord.size() == z_coord.size() ) )
    throw std::runtime_error ( "*** Error: The coordinate arrays must be of the same size." );

  if ( !this->x_coord.empty())
  {
    std::cout << "*** Warning: Previously registered coordinates will be erased,"
	      << " together with any attributes." << std::endl;
    this->x_coord.clear();
    this->y_coord.clear();
    this->z_coord.clear();
  }
  
  if ( !x_coord.empty() )
  {
    this->x_coord = x_coord;
    this->y_coord = y_coord;
    this->z_coord = z_coord;
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_vertex_marker ( unsigned int marker, const std::vector<unsigned int> &vertices )
{
  if (marker != 0)
  {
    for ( std::vector<unsigned int>::const_iterator it = vertices.begin();
	  it != vertices.end(); ++it )
    {
      vertex_markers[ ( *it ) ] = marker;
    }
    plc_builded = false;
  }
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet (const std::vector<unsigned int> &polygon, unsigned int marker, double max_area)
{
  std::vector< std::vector<unsigned int> > facet;
  facet.push_back(polygon);
  facet_list.push_back(facet);
  
  // Add a facet marker
  if (marker != 0)
  {
    add_facet_marker(marker);
    if (max_area != -1.0)
    {
      add_facet_area_constraints(max_area);
    }
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_polygon_to_last_facet(const std::vector<unsigned int> &polygon)
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    std::vector< std::vector<unsigned int> > &facet = facet_list.back();
    facet.push_back(polygon);
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_polygon_to_last_facet(const std::vector<unsigned int> &polygon, const std::vector<double> &point)
{
  add_polygon_to_last_facet(polygon);
  add_facet_hole(point);
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_polygon_to_facet(const std::vector<unsigned int> &polygon, unsigned int facet_ind, bool as_a_hole)
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else if (!facet_list.size() > facet_ind)
    throw std::runtime_error ( "*** Error: Facet index is not registered yet." );
  else 
  {
    std::vector< std::vector<unsigned int> > &facet = facet_list[facet_ind];
    facet.push_back(polygon);
    if (as_a_hole)
    {
      if ((*max_element(polygon.begin(),polygon.end())) > x_coord.size())
	throw std::runtime_error ( "*** Error: Vertex number in polygon can't exceed the number of registered vertices." );
      double x_mean = 0;
      double y_mean = 0;
      double z_mean = 0;
      unsigned int num_vert = polygon.size();
      // Iterate over the open polygon and collect the mean of the coordinates
      for (unsigned int i = 0; i < num_vert; i++)
      {
	x_mean += x_coord[polygon[i]];
	y_mean += y_coord[polygon[i]];
	z_mean += z_coord[polygon[i]];
      }
      x_mean /= num_vert;
      y_mean /= num_vert;
      z_mean /= num_vert;
  
      std::vector<double> point;
      point.push_back(x_mean);
      point.push_back(y_mean);
      point.push_back(z_mean);
  
      add_facet_hole(point, facet_ind);
	  
    }
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet_hole (const std::vector<double> &point)
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    // The last facet index
    unsigned int ind_last_facet = facet_list.size()-1;
    // Are there any holes registered for this facet before?
    if (facet_holes.find(ind_last_facet) == facet_holes.end())
      facet_holes[ind_last_facet] = point;
    else
      facet_holes[ind_last_facet].insert(facet_holes[ind_last_facet].end(), point.begin(), point.end());
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet_hole (const std::vector<double> &point, unsigned int facet_ind)
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    // Are there any holes registered for this facet before?
    if (facet_holes.find(facet_ind) == facet_holes.end())
      facet_holes[facet_ind] = point;
    else
      facet_holes[facet_ind].insert(facet_holes[facet_ind].end(), point.begin(), point.end());
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet_marker ( unsigned int marker )
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    facet_markers[facet_list.size()-1] = marker;
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet_marker ( unsigned int marker, unsigned int facet_ind )
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    facet_markers[facet_ind] = marker;
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet_area_constraints( double max_area )
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    if (facet_markers.find(facet_list.size()-1) == facet_markers.end())
      throw std::runtime_error ( "*** Error: Register a marker to the facet first.");
    else
      facet_area_constraints[facet_markers[facet_list.size()-1]] = max_area;
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_facet_area_constraints( double max_area, unsigned int facet_ind)
{
  if (!facet_list.size() > 0)
    throw std::runtime_error ( "*** Error: Register a facet first." );
  else
  {
    if (facet_markers.find(facet_ind) == facet_markers.end())
      throw std::runtime_error ( "*** Error: Register a marker to the facet first.");
    else
      facet_area_constraints[facet_markers[facet_ind]] = max_area;
  }
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_hole ( const std::vector<double> &point )
{
  hole_list.insert(hole_list.begin(), point.begin(), point.end());
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::add_region ( const std::vector<double> &point, double attribute, double max_volume)
{
  region_list.insert(region_list.end(), point.begin(), point.end());
  region_list.push_back(attribute);
  region_list.push_back(max_volume);
  plc_builded = false;
}
//-----------------------------------------------------------------------------
void PLCEditor::build_plc()
{
  plc.clear();
  // Add vertices
  unsigned int  num_vertices = x_coord.size();
  if (!num_vertices > 0)
    throw std::runtime_error ( "*** Error: Register vertices before build." );
  else
  {
    plc.data.numberofpoints = num_vertices;
    plc.data.pointlist = new double[num_vertices * 3];
    unsigned int ind = 0;
    for (unsigned int i = 0; i < num_vertices; ++i)
    {
      plc.data.pointlist[ind++] = x_coord[i];
      plc.data.pointlist[ind++] = y_coord[i];
      plc.data.pointlist[ind++] = z_coord[i];
    }
  }
      
  // Add vertex marker
  if (vertex_markers.size() > 0)
  {
    plc.data.pointmarkerlist = new int[num_vertices];
      
    // Initilize the pointmarkerlist with zeros
    for (unsigned int i = 0; i < num_vertices; ++i)
      plc.data.pointmarkerlist[i] = 0;
      
    // Iterate through the vertex marker map and add the marker
    for (std::map<unsigned int,unsigned int>::const_iterator it = vertex_markers.begin();
	 it != vertex_markers.end(); ++it)
      plc.data.pointmarkerlist[(*it).first] = (*it).second;
  }
      
  // Add facets
  const unsigned int num_facets = facet_list.size();
  if (!num_facets > 0)
    throw std::runtime_error ( "*** Error: Register facets before build." );
  else
  {
    tetgen::tetgenio::facet *f;   // Pointer to a facet
    tetgen::tetgenio::polygon *p; // Pointer to a polygon
      
    // Create the array of facets
    plc.data.numberoffacets = num_facets;
    plc.data.facetlist = new tetgen::tetgenio::facet[num_facets];
      
    for (unsigned int facet_ind = 0; facet_ind < num_facets; ++facet_ind)
    {
      // Reference to the present facet
      f = &plc.data.facetlist[facet_ind];
      
      // Register any facet holes
      if (facet_holes.find(facet_ind) == facet_holes.end())
      {
	f->numberofholes = 0;
	f->holelist = NULL;
      }
      else
      {
	f->numberofholes = facet_holes[facet_ind].size() / 3;
	f->holelist = new double[facet_holes[facet_ind].size()];
	for (unsigned int  i = 0; i < facet_holes[facet_ind].size(); ++i)
	  f->holelist[i] = facet_holes[facet_ind][i];
      }
      
      // Create the array of polygons
      const unsigned int  num_polygons = facet_list[facet_ind].size();
      f->numberofpolygons = num_polygons;
      f->polygonlist = new tetgen::tetgenio::polygon[num_polygons];
      
      for (unsigned int  polygon_ind = 0; polygon_ind < num_polygons; ++polygon_ind)
      {
	// Reference to the present polygon
	p = &f->polygonlist[polygon_ind];
      
	// Create the array of vertices
	const unsigned int  num_vertices = facet_list[facet_ind][polygon_ind].size();
	p->numberofvertices = num_vertices;
	p->vertexlist = new int[num_vertices];
      
	// Fill the polygon with vertices
	for (unsigned int  vertice_ind = 0; vertice_ind < num_vertices; ++vertice_ind)
	  p->vertexlist[vertice_ind] = facet_list[facet_ind][polygon_ind][vertice_ind];
      }
    }
  }
      
  // Add facet markers
  if (facet_markers.size() > 0)
  {
    plc.data.facetmarkerlist = new int[num_facets];
      
    // Initilize the facetmarkerlist with zeros
    for (unsigned int  i = 0; i < num_facets; ++i)
      plc.data.facetmarkerlist[i] = 0;
      
    // Iterate through the facet marker map and add the marker
    for (std::map<unsigned int,unsigned int>::const_iterator it = facet_markers.begin();
	 it != facet_markers.end(); ++it)
      plc.data.facetmarkerlist[(*it).first] = (*it).second;
  }
  
  // Add facet area constraints
  if (facet_area_constraints.size() > 0)
  {
    plc.data.numberoffacetconstraints = facet_area_constraints.size();
    plc.data.facetconstraintlist = new double[facet_area_constraints.size()*2];
    unsigned int  i = 0;
    for (std::map<unsigned int,double>::const_iterator it = facet_area_constraints.begin();
	 it != facet_area_constraints.end(); ++it)
    {
      plc.data.facetconstraintlist[i++] = static_cast<double>((*it).first);
      plc.data.facetconstraintlist[i++] = (*it).second;
    }
  }
      
  // Add holes
  if (hole_list.size() > 0)
  {
    plc.data.numberofholes = hole_list.size() / 3;
    plc.data.holelist = new double[hole_list.size()];
    for (unsigned int  i = 0; i < hole_list.size(); ++i){
      plc.data.holelist[i] = hole_list[i];}
  }
      
  // Add regions
  if (region_list.size() > 0)
  {
    plc.data.numberofregions = region_list.size() / 5;
    plc.data.regionlist = new double[region_list.size()];
    for (unsigned int  i = 0; i < region_list.size(); ++i)
    {
      plc.data.regionlist[i] = region_list[i];
      // If the ith element in region_list refere to a regional attribute
      if (i % 5 == 3)
      {
	if (region_list[i] != -1.0)
	{
	  plc.use_regional_attributes = true;
	}
      }
      // If the ith element in region_list refere to a regional volume constraint
      if (i % 5 == 4)
      {
	if (region_list[i] != -1.0)
	{
	  plc.use_regional_volume_constraints = true;
	}
      }
    }
  }
  plc_builded = true;
}
//-----------------------------------------------------------------------------
void PLCEditor::build(std::string file_name, bool save_tetgen, bool save_off)
{
  build();
  save_mesh(file_name, save_tetgen, save_off);
}
//-----------------------------------------------------------------------------
void PLCEditor::build()
{
  if (!plc_builded)
  {
    build_plc();
  }
  plc.build_mesh(tetmesh);
  tetmesh_builded = true;
}
//-----------------------------------------------------------------------------
void PLCEditor::save_mesh(std::string file_name, bool save_tetgen, bool save_off)
{
  if (!tetmesh_builded)
  {
    throw std::runtime_error ( "*** Error: The mesh is not built." );
  }
  if (save_tetgen)
  {
    tetmesh.save_tetgen(file_name);
  }
  if (save_off)
  {
    tetmesh.save_off(file_name);
  }
  tetmesh.save_dolfin(file_name);
}
//-----------------------------------------------------------------------------
void PLCEditor::clear()
{
  x_coord.clear();
  y_coord.clear();
  z_coord.clear();
  vertex_markers.clear();
  
  facet_list.clear();
  facet_holes.clear();
  facet_markers.clear();
  facet_area_constraints.clear();
  
  hole_list.clear();
  region_list.clear();
  plc.clear();
  tetmesh.clear();
}
//-----------------------------------------------------------------------------
void PLCEditor::test_plc()
{
  if (!plc_builded)
  {
    build_plc();
    plc_builded = true;
  }
  plc.test();
}
//-----------------------------------------------------------------------------
