// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include "tetmesh.h"
#include <tetgen_mesh.h>

#include <sstream>
#include <iostream>
#include <string>
#include <stdexcept>
#include <vector>
#include <algorithm>

#include <dolfin/io/File.h>
#include <dolfin/mesh/MeshEditor.h>
#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/MeshTopology.h>
#include <dolfin/mesh/MeshFunction.h>
#include <dolfin/mesh/MeshConnectivity.h>

using namespace tritetmesh;

//-----------------------------------------------------------------------------
TetMesh::TetMesh()
{}
//-----------------------------------------------------------------------------
TetMesh::~TetMesh()
{}
//-----------------------------------------------------------------------------
void TetMesh::load_tetgen(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  
  clear();
  data.load_tetmesh(file_name);
}
//-----------------------------------------------------------------------------
void TetMesh::save_tetgen(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  data.save_nodes(file_name);
  data.save_faces(file_name);
  data.save_elements(file_name);
}
//-----------------------------------------------------------------------------
tetgen::tetgenmesh *TetMesh::reconstruct_mesh()
{
  tetgen_behavior.parse_commandline((char *)"r");
  
  tetgen::tetgenmesh *m;
  m = new tetgen::tetgenmesh();
  
  m->in = &data;
  m->b = &tetgen_behavior;
  m->initializepools();
  m->transfernodes();  
  m->reconstructmesh();
  return m;
}
//-----------------------------------------------------------------------------
void TetMesh::save_medit(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  
  tetgen::tetgenmesh *m;
  m = reconstruct_mesh();

  m->outmesh2medit(file_name);
  delete m;
}
//-----------------------------------------------------------------------------
void TetMesh::save_gid(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  tetgen::tetgenmesh *m;
  m = reconstruct_mesh();

  m->outmesh2gid(file_name);
  delete m;
}
//-----------------------------------------------------------------------------
void TetMesh::save_off(std::string in_file_name)
{
  char file_name[1024];
  strcpy(file_name, in_file_name.c_str());
  
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  tetgen::tetgenmesh *m;
  m = reconstruct_mesh();

  m->outmesh2off(file_name);
  delete m;
}
//-----------------------------------------------------------------------------
void TetMesh::clear()
{
  data.deinitialize();
  data.initialize();
}
//-----------------------------------------------------------------------------
void TetMesh::check_consistency()
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to check consistency for." );
  tetgen::tetgenmesh *m;
  m = reconstruct_mesh();

  m->checkmesh();
  delete m;
}

//-----------------------------------------------------------------------------
const TetMesh& TetMesh::operator=(const TetMesh& tetmesh)
{
  data = tetmesh.data;
  return *this;
}
//-----------------------------------------------------------------------------
void TetMesh::refine_mesh(TetMesh &output, std::string in_switches)
{
  char switches[1024];
  strcpy(switches, in_switches.c_str());

  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to refine." );

  /* FIXME: Implement variable volume refinement, or try to do it upstream in tetgen.cxx?
  // Any volume functions registered?
   
  if (volume_functions.size()>0)
    {
      bool refine_more = true;
      while (refine_more)
	{
	  
	}
    }
  */
  output.clear();
  tetgen::tetrahedralize(switches, &data, &output.data);
}
//-----------------------------------------------------------------------------
void TetMesh::save_dolfin(std::string file_name)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  dolfin::MeshEditor editor;
  dolfin::Mesh dolfin_mesh;
  std::cout << "mesh name: " << file_name + ".xml" << std::endl;
  dolfin::File mesh_file(file_name + ".xml");
  
  std::cout << "Saving mesh to " << file_name << ".xml" << std::endl;
  
  editor.open(dolfin_mesh,"tetrahedron",3,3);
  editor.init_vertices(data.numberofpoints);
  editor.init_cells(data.numberoftetrahedra);
  
  // Add vertex coordinates to the dolfin mesh
  for (int i = 0; i < data.numberofpoints; i++)
    {
      editor.add_vertex(i,data.pointlist[i*3],data.pointlist[i*3+1],data.pointlist[i*3+2]);
    }
      
  // Add cell information to the dolfin mesh 
  for (int i = 0; i < data.numberoftetrahedra; i++)
    {
      editor.add_cell(i, data.tetrahedronlist[i*4], data.tetrahedronlist[i*4+1],
		      data.tetrahedronlist[i*4+2], data.tetrahedronlist[i*4+3]);
    }
  editor.close();
  
  mesh_file << dolfin_mesh;

  // Any vertex markers to save?
  if (!(data.pointmarkerlist == 0))
    {
      std::cout << "Saving vertex markers to " << file_name
		<< "_vertex_markers.xml" << std::endl;
      dolfin::MeshFunction<unsigned int> vertex_markers(dolfin_mesh,0);
      for (int i = 0; i < data.numberofpoints; i++)
	{
	  vertex_markers[i] = data.pointmarkerlist[i];
	}
      dolfin::File vertex_marker_file(file_name + "_vertex_markers.xml");
      vertex_marker_file << vertex_markers;
    }
  
  // Any face markers to save? 
  if (!(data.trifacemarkerlist == 0))
    {
      // Tetgen only saves the faces on the boundary together with any marker.
      // A 2D MeshFunction is defined over all triangles and we therefore need
      // to map the triangles information we get from tetgen to the information we
      // are saving in the MeshFunction.
      std::cout << "Saving face markers to " << file_name
		<< "_face_markers.xml" << std::endl;
      dolfin_mesh.init(0,2);
      dolfin_mesh.init(2,3);
      dolfin::MeshTopology& top = dolfin_mesh.topology();
      dolfin::MeshConnectivity& con02 = top(0,2);
      dolfin::MeshConnectivity& con23 = top(2,3);
      
      dolfin::MeshFunction<unsigned int> face_markers(dolfin_mesh,2);

      // Initialize the MeshFunction
      face_markers = 0;
      
      // Allocate memory
      std::vector<unsigned int> faces0,faces1,faces2,common_faces0,common_faces1;
      int enough = 100;
      faces0.reserve(enough);
      faces1.reserve(enough);
      faces2.reserve(enough);
      common_faces0.reserve(enough);
      common_faces1.reserve(enough);
      int i;
      unsigned int vertex0,vertex1,vertex2, j, face;
      std::vector<unsigned int>::iterator it0, it1;

      // Iterate over the triangles on the boundary and find the
      // global index. Use this and register the marker
      for (i = 0; i < data.numberoftrifaces; i++)
	{
	  // Only need to search for the face if the marker is different from 0
	  if (data.trifacemarkerlist[i] != 0)
	    {
	      // The vertex numbers in the i:th face
	      vertex0 = data.trifacelist[i*3];
	      vertex1 = data.trifacelist[i*3+1];
	      vertex2 = data.trifacelist[i*3+2];
	      
	      // The face numbers for each vertex
	      for (j = 0; j < con02.size(vertex0); j++)
		{
		  face = con02(vertex0)[j];
		  // If the face is on the boundary
		  if (!(con23.size(face)>1))
		    {
		      // Register the face
		      faces0.push_back(face);
		    }
		}
	      // The face numbers for each vertex
	      for (j = 0; j < con02.size(vertex1); j++)
		{
		  face = con02(vertex1)[j];
		  // If the face is on the boundary
		  if (!(con23.size(face)>1))
		    {
		      // Register the face
		      faces1.push_back(face);
		    }
		}
	      // The face numbers for each vertex
	      for (j = 0; j < con02.size(vertex2); j++)
		{
		  face = con02(vertex2)[j];
		  // If the face is on the boundary
		  if (!(con23.size(face)>1))
		    {
		      // Register the face
		      faces2.push_back(face);
		    }
		}
	      
	      // Sort the triangle indices (set_intersection need a sorted vector)
	      std::sort(faces0.begin(),faces0.end());
	      std::sort(faces1.begin(),faces1.end());
	      std::sort(faces2.begin(),faces2.end());
	      
	      // Find the triangle index that is common for all three vertices
	      it0 = std::set_intersection (faces0.begin(),faces0.end(),
					   faces1.begin(),faces1.end(),
					   common_faces0.begin());
	      
	      it1 = std::set_intersection (common_faces0.begin(),it0,
					   faces2.begin(),faces2.end(),
					   common_faces1.begin());
	      if (int(it1 - common_faces1.begin()) == 0)
		{
		  throw std::runtime_error ( "*** Error: No common faces." );
		}
	      if (int(it1 - common_faces1.begin()) > 1)
		{
		  throw std::runtime_error ( "*** Error: To many common faces." );
		}
	      face_markers[common_faces1[0]] = data.trifacemarkerlist[i];
	    }
	  faces0.clear();
	  faces1.clear();
	  faces2.clear();
	  common_faces0.clear();
	  common_faces1.clear();
	}
      dolfin::File face_marker_file(file_name + "_face_markers.xml");
      face_marker_file << face_markers;
    }
  
  // Any tetrahedron attributes to save?
  if (!(data.numberoftetrahedronattributes == 0))
    {
      std::cout << "Saving tetrahedron attributes to " << file_name
		<< "_tetrahedron_attributes.xml" << std::endl;
      dolfin::MeshFunction<double> tetrahedron_attributes(dolfin_mesh,3);
      for (int i = 0; i < data.numberoftetrahedra; i++)
	{
	  tetrahedron_attributes[i] = data.tetrahedronattributelist[i];
	}
      dolfin::File tetrahedron_attribute_file(file_name + "_tetrahedron_attributes.xml");
      tetrahedron_attribute_file << tetrahedron_attributes;
    }
}
