// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include "pslg.h"

#include <stdexcept>
#include <iostream>
#include <sstream>
#include <string>
#include <cstring>
#include <cstdlib>
#include <cstdio>

using namespace tritetmesh;

//-----------------------------------------------------------------------------
PSLG::PSLG():
  minimum_angle(-1.0),	       
  uniform_area_constraint(-1.0),
  verbosity(1),
  use_regional_attributes(false),
  use_regional_area_constraints(false),
  generate_edge_information(false),
  enclose_convex_hull(false),
  area_func(0),
  data()
{
  initialize();
}
//-----------------------------------------------------------------------------
PSLG::~PSLG()
{
  deinitialize();
}
//-----------------------------------------------------------------------------
void PSLG::initialize()
{
  data.pointlist  = (double *) NULL;
  data.pointattributelist = (double *) NULL;
  data.pointmarkerlist = (int *) NULL;
  data.numberofpoints = 0;
  data.numberofpointattributes = 0;

  data.trianglelist = (int *) NULL;
  data.triangleattributelist = (double *) NULL;
  data.trianglearealist = (double *) NULL;
  data.neighborlist = (int *) NULL;
  data.numberoftriangles = 0;
  data.numberofcorners = 0;
  data.numberoftriangleattributes = 0;

  data.segmentlist = (int *) NULL;
  data.segmentmarkerlist = (int *) NULL;
  data.numberofsegments = 0;

  data.holelist = (double *) NULL;
  data.numberofholes = 0;

  data.regionlist = (double *) NULL;
  data.numberofregions = 0;      

  data.edgelist = (int *) NULL;
  data.edgemarkerlist = (int *) NULL;
  data.normlist = (double *) NULL;
  data.numberofedges = 0;
}
//-----------------------------------------------------------------------------
void PSLG::deinitialize()
{ 
  if (data.pointlist != (double *) NULL) {
    delete [] data.pointlist;
  }
  if (data.pointattributelist != (double *) NULL) {
    delete [] data.pointattributelist;
  }
  if (data.pointmarkerlist != (int *) NULL) {
    delete [] data.pointmarkerlist;
  }
  if (data.segmentlist != (int *) NULL) {
    delete [] data.segmentlist;
  }
  if (data.segmentmarkerlist != (int *) NULL) {
    delete [] data.segmentmarkerlist;
  }
  if (data.holelist != (double *) NULL) {
    delete [] data.holelist;
    data.holelist = (double *) NULL;
  }
  if (data.regionlist != (double *) NULL) {
    delete [] data.regionlist;
  }
}
//-----------------------------------------------------------------------------
const PSLG& PSLG::operator=(const PSLG& pslg)
{
  data = pslg.data;
  return *this;
}
//-----------------------------------------------------------------------------
void PSLG::print_triangle_parameters()
{
  std::cout << std::endl;
  std::cout << "Triangle parameters: " << std::endl;
  std::cout << "+----------------------------------------------" << std::endl;     
  std::cout << "| Minmum angle                    (q) : ";
  if (minimum_angle == -1) std::cout << "Not used"  << std::endl; else std::cout << minimum_angle << std::endl;
  std::cout << "| Uniform area constraint      (aX.X) : ";
  if (uniform_area_constraint == -1) std::cout << "Not used"  << std::endl; else std::cout << uniform_area_constraint << std::endl;
  std::cout << "| Verbosity                       (V) : " << verbosity << std::endl;
  std::cout << "| Use regional attributes         (A) : " << (use_regional_attributes  ? "True": "False") << std::endl;
  std::cout << "| Use regional area constraints   (a) : " << (use_regional_area_constraints  ? "True": "False") << std::endl;
  std::cout << "| Use user defined area function  (u) : " << (area_func !=0 ? "True" : "False" ) << std::endl;
  std::cout << "| Generate edgeinformation        (e) : " << (generate_edge_information  ? "True": "False") << std::endl;
  std::cout << "| Enclose a convex hull with seg. (c) : " << (enclose_convex_hull  ? "True": "False") << std::endl;
  std::cout << "+----------------------------------------------" << std::endl << std::endl;     
}
//-----------------------------------------------------------------------------
// FIXME: Remove when proved not needed
void PSLG::build_mesh(TriMesh &output, char* switches)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: PSLG empty. Nothing to build." );
  output.clear();
  generate(output,switches);
}
//-----------------------------------------------------------------------------
void PSLG::build_mesh(TriMesh &output)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: PSLG empty. Nothing to build." );
  output.clear();
  
  /// Build the switch string
  std::string switches("pz");

  if (use_regional_attributes)
    {
      switches += "A";
    }

  if (use_regional_area_constraints)
    {
      switches += "a";
    }
  
  if (enclose_convex_hull)
    {
      switches += "c";
    }
  
  if (minimum_angle > 1.0) 
    {
      switches += "q";
      std::ostringstream o;
      if (!(o << minimum_angle))
	throw std::runtime_error ("Couldn't convert double to string");
      switches += o.str();
    }
  
  if (uniform_area_constraint > 0.0)
    {
      switches += "a";
      std::ostringstream o;
      if (!(o << uniform_area_constraint))
	throw std::runtime_error ("Couldn't convert double to string");
      switches += o.str();
    }
  
   if (area_func != 0)
    {
      switches += "u";
      // Set the global area_function pointer to the one registered in the trimesh
      global_area_function = area_func;
    }

   if ( 0 <= verbosity && verbosity <= 3)
     {
       if (!(verbosity == 1))
	{
	  if (verbosity == 0) 
	    {
	      switches += "Q";
	    }
	  else 
	    for (int i = 1; i <= verbosity; i++ )
	      switches += "V";
	}
     }
   
   // Convert the const char* to a char* 
   char cstr_switches[switches.size()];
   std::strcpy(cstr_switches,switches.c_str());
   std::cout << "Triangle switches: " << cstr_switches << std::endl;
   
   generate(output,cstr_switches);
   global_area_function = 0;
}
//-----------------------------------------------------------------------------
void PSLG::generate(TriMesh &output, char* switches)
{
  
  if (verbosity > 0)
    {
      std::cout << "+----------------------------------------------------+" << std::endl;
      std::cout << "| Generating mesh from a Planar Straight Line Graph. |" << std::endl;
      std::cout << "+----------------------------------------------------+" << std::endl << std::endl;
      print_triangle_parameters();
    }
  
  triangulate(switches, &data, &output.data, (struct triangulateio *) NULL);

  // This makes it easier to delete the objetcs...
  output.data.regionlist = (double * ) NULL;
  output.data.holelist = (double * ) NULL;
}
//-----------------------------------------------------------------------------
void PSLG::save_triangle(char* file_name)
{
  FILE *fout;
  char outfilename[FILENAMESIZE];
  int i, j;

  // Save the nodes with any attributes or markers
  sprintf(outfilename, "%s.node", file_name);
  printf("Saving nodes to %s\n", outfilename);
  fout = fopen(outfilename, "w");
  fprintf(fout, "%d  %d  %d  %d\n", data.numberofpoints, 2,
          data.numberofpointattributes, data.pointmarkerlist != NULL ? 1 : 0);
  for (i = 0; i < data.numberofpoints; i++) 
    {
      fprintf(fout, "%d  %.16g  %.16g", i , data.pointlist[i * 2],
              data.pointlist[i * 2 + 1]);
      for (j = 0; j < data.numberofpointattributes; j++) 
	{
	  fprintf(fout, "  %.16g", 
		  data.pointattributelist[i * data.numberofpointattributes + j]);
	}
      if (data.pointmarkerlist != NULL) 
	{
	  fprintf(fout, "  %d", data.pointmarkerlist[i]);
	}
      fprintf(fout, "\n");
    }
  fclose(fout);

  // Save the input segments
  sprintf(outfilename, "%s.poly", file_name);
  printf("Saving segments to %s\n", outfilename);
  fout = fopen(outfilename, "w");
  fprintf(fout, "%d  %d  %d  %d\n", 0, 2, data.numberofpointattributes,
          data.pointmarkerlist != NULL ? 1 : 0);
  fprintf(fout, "%ld  %d\n", static_cast<long int>(data.numberofsegments), 
	  data.segmentmarkerlist != NULL ? 1 : 0);
  for (i = 0; i < data.numberofsegments; i++) 
    {
      fprintf(fout, "%d  %d  %d", i , data.segmentlist[i * 2],
              data.segmentlist[i * 2 + 1]);
      if (data.segmentmarkerlist != NULL) 
	{
	  fprintf(fout, "  %d", data.segmentmarkerlist[i]);
	}
      fprintf(fout, "\n");
    }

  // Save holes.
  fprintf(fout, "%d\n", data.numberofholes);
  for (i = 0; i < data.numberofholes; i++) {
    // Output x, y coordinates.
    fprintf(fout, "%d  %.12g  %.12g", i, data.holelist[i * 2],
            data.holelist[i * 2 + 1]);
    fprintf(fout, "\n");
  }

  // Save regions.
  fprintf(fout, "%d\n", data.numberofregions);
  for (i = 0; i < data.numberofregions; i++) {
    // Output the index, x, y coordinates, attribute (region number)
    //   and maximum area constraint (maybe -1).
    fprintf(fout, "%d  %.12g  %.12g  %.12g  %.12g\n", i,
	    data.regionlist[i * 4], data.regionlist[i * 4 + 1],
	    data.regionlist[i * 4 + 2], data.regionlist[i * 4 + 3]);
  }
  fclose(fout);
}
//-----------------------------------------------------------------------------
