// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include "pslgeditor.h"
#include "pslg.h"

#include <cstring>
#include <stdexcept>
#include <cstdlib>
#include <stdio.h>
#include <iostream>
#include <algorithm>

using namespace tritetmesh;

//-----------------------------------------------------------------------------
PSLGEditor::PSLGEditor():
  pslg_builded(false)
{}
//-----------------------------------------------------------------------------
PSLGEditor::~PSLGEditor()
{}
//-----------------------------------------------------------------------------
void PSLGEditor::print_vertices()
{
  if (x_coord.size() == 0)
    {
      std::cout << "*** Warning: No vertices registrated." << std::endl;
    }
  for (unsigned int i = 0; i < x_coord.size(); ++i)
    {
      std::cout << "Vertex " << i << ": (" << x_coord[i] << ", " 
		<< y_coord[i] << ")"<< std::endl;
    }
}
//-----------------------------------------------------------------------------
void PSLGEditor::register_vertices(const std::vector<double> &x_coord,
		       const std::vector<double> &y_coord)
{
  if ( ! ( x_coord.size() == y_coord.size() ))
    throw std::runtime_error ( "*** Error: The coordinate sequences must be of the same size." );
  if ( !this->x_coord.empty())
    {
      std::cout << "*** Warning: Previously registered coordinates will be erased,"
                << " together with any attributes." << std::endl;
      this->x_coord.clear();
      this->y_coord.clear();
      vertex_attributes.clear();
    }
  if ( !x_coord.empty() )
    {
      this->x_coord = x_coord;
      this->y_coord = y_coord;
    }
}
//-----------------------------------------------------------------------------
void PSLGEditor::register_vertices(const std::vector<double> &x_coord,
				   const std::vector<double> &y_coord,
				   const std::vector<double> &attributes)
{
  register_vertices(x_coord,y_coord);
  add_vertex_attributes(attributes);
}
//-----------------------------------------------------------------------------
void PSLGEditor::register_vertices(const std::vector<double> &x_coord,
				   const std::vector<double> &y_coord,
				   const std::vector<double> &attributes1,
				   const std::vector<double> &attributes2)
{
  register_vertices(x_coord,y_coord);
  add_vertex_attributes(attributes1);
  add_vertex_attributes(attributes2);
}
//-----------------------------------------------------------------------------
void PSLGEditor::register_vertices(const std::vector<double> &x_coord,
				   const std::vector<double> &y_coord,
				   const std::vector<double> &attributes1,
				   const std::vector<double> &attributes2,
				   const std::vector<double> &attributes3)
{
  register_vertices(x_coord,y_coord);
  add_vertex_attributes(attributes1);
  add_vertex_attributes(attributes2);
  add_vertex_attributes(attributes3);
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_vertex_attributes ( const std::vector<double> & attributes)
{
  if ( ! ( x_coord.size() == attributes.size() ))
    throw std::runtime_error ( "*** Error: The number of attributes must be of the same as the number of registered vertices." );
  
  vertex_attributes.push_back(attributes);
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_vertex_marker ( unsigned int marker, const std::vector<unsigned int> &vertices )
{
  for ( std::vector<unsigned int>::const_iterator it = vertices.begin();
	it != vertices.end(); ++it )
    {
      vertex_markers[ ( *it ) ] = marker;
    }
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_polygon (const std::vector<unsigned int> &polygon, double attribute, double max_area)
{
  if (polygon.size() < 3)
    throw std::runtime_error ( "*** Error: The number of vertices provided needs to be 3 or more." );
  // Check if the polygon is closed
  bool closed = polygon.front() == polygon.back();
  if (!closed)
    {
      // Copy the input vector and close the polygon
      std::vector<unsigned int> closed_polygon = polygon;
      closed_polygon.push_back(polygon.front());
      add_segments(closed_polygon);
    }
  else
    {
      add_segments(polygon);
    }

  if (attribute != 0.0 or max_area > 0.0)
    {
      double x_mean = 0;
      double y_mean = 0;
      unsigned int num_vert = polygon.size() - (closed ? 1 : 0);
      // Iterate over the open polygon and collect the mean of the coordinates
      for (unsigned int i = 0; i < num_vert; i++)
	{
	  x_mean += x_coord[i];
	  y_mean += y_coord[i];
	}
      x_mean /= num_vert;
      y_mean /= num_vert;

      std::vector<double> point;
      point.push_back(x_mean);
      point.push_back(y_mean);

      add_region(point, attribute, max_area);
    }
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_polygon_hole (const std::vector<unsigned int> &polygon)
{
  if (polygon.size() < 3)
    throw std::runtime_error ( "*** Error: The number of vertices provided needs to be 3 or more." );
  // Check if the polygon is closed
  bool closed = polygon.front() == polygon.back();
  if (!closed)
    {
      // Copy the input vector and close the polygon
      std::vector<unsigned int> closed_polygon = polygon;
      closed_polygon.push_back(polygon.front());
      add_segments(closed_polygon);
    }
  else
    {
      add_segments(polygon);
    }

  double x_mean = 0;
  double y_mean = 0;
  unsigned int num_vert = polygon.size() - (closed ? 1 : 0);
  // Iterate over the open polygon and collect the mean of the coordinates
  for (unsigned int i = 0; i < num_vert; i++)
    {
      x_mean += x_coord[polygon[i]];
      y_mean += y_coord[polygon[i]];
    }
  x_mean /= num_vert;
  y_mean /= num_vert;
  
  std::vector<double> point;
  point.push_back(x_mean);
  point.push_back(y_mean);
  
  add_hole(point);
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_segments(const std::vector<unsigned int> &segments, unsigned int marker)
{
  if ((*max_element(segments.begin(),segments.end())) > x_coord.size())
    throw std::runtime_error ( "*** Error: Vertex numbers can't exceed the number of registered vertices." );
  if (segments.size() < 2)
    throw std::runtime_error ( "*** Error: The number of vertices provided needs to be 2 or more." );
  unsigned int num_seg = segments.size()-1;
  for (unsigned int i = 0; i < num_seg; i++)
    {
      // Add the segments as a pair. Keep the first element the smallest.
      int smallest = (segments[i] < segments[i+1] ? segments[i]:segments[i+1]);
      int largest  = (segments[i] > segments[i+1] ? segments[i]:segments[i+1]);
      if (smallest == largest)
	throw std::runtime_error ( "*** Error: Provide different vertices for a segment." );
      std::pair<unsigned int,unsigned int> seg(smallest,largest);
      segment_list.insert(seg);
      if (marker != 0)
	{
	  segment_markers[seg] = marker;
	}
    }
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_hole ( const std::vector<double> &point )
{
  hole_list.insert(hole_list.begin(), point.begin(), point.end());
}
//-----------------------------------------------------------------------------
void PSLGEditor::add_region ( const std::vector<double> &point, double attribute, double max_volume)
{
  region_list.insert(region_list.end(), point.begin(), point.end());
  region_list.push_back(attribute);
  region_list.push_back(max_volume);
}
//-----------------------------------------------------------------------------
void PSLGEditor::build_pslg()
{
  pslg.clear();
  // Add vertices
  unsigned int num_vertices = x_coord.size();
  if (!num_vertices > 0)
    throw std::runtime_error ( "*** Error: Register vertices before build." );
  else
    {
      pslg.data.numberofpoints = num_vertices;
      pslg.data.pointlist = new double[num_vertices * 3];
      unsigned int ind = 0;
      for (unsigned int i = 0; i < num_vertices; ++i)
	{
	  pslg.data.pointlist[ind++] = x_coord[i];
	  pslg.data.pointlist[ind++] = y_coord[i];
	}
    }

  // Add vertex attributes
  unsigned int num_vertex_attributes = vertex_attributes.size();
  if (num_vertex_attributes > 0)
    {
      pslg.data.numberofpointattributes = num_vertex_attributes;
      pslg.data.pointattributelist = new double[num_vertices*num_vertex_attributes];
      
      // Iterate through the vertex attribute vectors and add the value
      for (unsigned int i = 0; i < num_vertices; i++)
	{
	  for (unsigned int j = 0; j < num_vertex_attributes; j++)
	    {
	      pslg.data.pointattributelist[i*num_vertex_attributes + j] = vertex_attributes[j][i];
	    }
	}
    }
  
  // Add vertex marker
  if (vertex_markers.size() > 0)
    {
      pslg.data.pointmarkerlist = new int[num_vertices];
      
      // Initilize the pointmarkerlist with zeros
      for (unsigned int i = 0; i < num_vertices; ++i)
	{
	  pslg.data.pointmarkerlist[i] = 0;
	}
      // Iterate through the vertex marker map and add the marker
      for (std::map<unsigned int,unsigned int>::const_iterator it = vertex_markers.begin();
	   it != vertex_markers.end(); ++it)
	{
	  pslg.data.pointmarkerlist[(*it).first] = (*it).second;
	}
    }

  // Add the segments
  const unsigned int num_segments = segment_list.size();
  
  if (num_segments > 0)
    {
      pslg.data.segmentlist = new int[num_segments*2];
      pslg.data.numberofsegments = num_segments;
      
      if (segment_markers.size() > 0)
	{
	  // Initialize the segmentmarkerlist
	  pslg.data.segmentmarkerlist = new int[num_segments];
	  for (unsigned int i = 0; i<num_segments;i++)
	    {
	      pslg.data.segmentmarkerlist[i] = 0;
	    }
	}
      
      unsigned int i = 0;
      std::map<std::pair<unsigned int,unsigned int>,unsigned int >::iterator result;
      // Iterate over the segment pairs and add them to the segmentlist
      for (std::set<std::pair<unsigned int,unsigned int> >::const_iterator it = segment_list.begin();
	   it != segment_list.end(); it++)
	{
	  pslg.data.segmentlist[i*2]   = (*it).first;
	  pslg.data.segmentlist[i*2+1] = (*it).second;
	  
	  // Check for markers
	  result = segment_markers.find(*it);
	  if (result != segment_markers.end())
	    {
	      pslg.data.segmentmarkerlist[i] = (*result).second;
	    }
	  i++;
	}
    }
  
  // Add holes
  if (hole_list.size() > 0)
    {
      pslg.data.numberofholes = hole_list.size() / 2;
      pslg.data.holelist = new double[hole_list.size()];
      for (unsigned int i = 0; i < hole_list.size(); ++i)
	{
	  pslg.data.holelist[i] = hole_list[i];
	}
    }
  
  // Add regions
  if (region_list.size() > 0)
    {
      pslg.data.numberofregions = region_list.size() / 4;
      pslg.data.regionlist = new double[region_list.size()];
      for (unsigned int i = 0; i < region_list.size(); ++i)
	{
	  pslg.data.regionlist[i] = region_list[i];
	  // If the ith element in region_list refere to a regional attribute
	  if (i % 4 == 2)
	    {
	      if (region_list[i] != -1.0)
		{
		  pslg.use_regional_attributes = true;
		}
	    }
	  // If the ith element in region_list refere to a regional volume constraint
	  if (i % 4 == 3)
	    {
	      if (region_list[i] != -1.0)
		{
		  pslg.use_regional_area_constraints = true;
		}
	    }
	}
    }
  pslg_builded = true;
}
//-----------------------------------------------------------------------------
void PSLGEditor::build(std::string file_name, bool save_triangle)
{
  if (!pslg_builded)
    {
      build_pslg();
    }
  pslg.build_mesh(trimesh);
  trimesh.save_dolfin(file_name);
  if (save_triangle)
    {
      // Convert the const char* to a char* 
      char cstr_file_name[file_name.size()];
      std::strcpy(cstr_file_name,file_name.c_str());
      trimesh.save_triangle(cstr_file_name);
    }
}
//-----------------------------------------------------------------------------
void PSLGEditor::clear()
{
  x_coord.clear();
  y_coord.clear();
  
  vertex_attributes.clear();
  vertex_markers.clear();
  
  segment_list.clear();   
  segment_markers.clear();
  
  hole_list.clear();
  region_list.clear();

  pslg.clear();
  trimesh.clear();
  
  pslg_builded = false;
}
//-----------------------------------------------------------------------------
