// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include "trimesh.h"

#include <algorithm>
#include <string>
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstring>
#include <cstdlib>
#include <stdio.h>

#include <dolfin/io/File.h>
#include <dolfin/mesh/MeshEditor.h>
#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/MeshTopology.h>
#include <dolfin/mesh/MeshFunction.h>
#include <dolfin/mesh/MeshConnectivity.h>

using namespace tritetmesh;

//-----------------------------------------------------------------------------
TriMesh::TriMesh():
  minimum_angle(-1.0),	       
  uniform_area_constraint(-1.0),
  verbosity(1),
  area_func(0),
  data()
{
  initialize();
}
//-----------------------------------------------------------------------------
TriMesh::~TriMesh()
{
  deinitialize();
}
//-----------------------------------------------------------------------------
void TriMesh::initialize()
{
  data.pointlist  = (double *) NULL;
  data.pointattributelist = (double *) NULL;
  data.pointmarkerlist = (int *) NULL;
  data.numberofpoints = 0;
  data.numberofpointattributes = 0;

  data.trianglelist = (int *) NULL;
  data.triangleattributelist = (double *) NULL;
  data.trianglearealist = (double *) NULL;
  data.neighborlist = (int *) NULL;
  data.numberoftriangles = 0;
  data.numberofcorners = 3;
  data.numberoftriangleattributes = 0;

  data.segmentlist = (int *) NULL;
  data.segmentmarkerlist = (int *) NULL;
  data.numberofsegments = 0;

  data.holelist = (double *) NULL;
  data.numberofholes = 0;

  data.regionlist = (double *) NULL;
  data.numberofregions = 0;      

  data.edgelist = (int *) NULL;
  data.edgemarkerlist = (int *) NULL;
  data.normlist = (double *) NULL;
  data.numberofedges = 0;
}
//-----------------------------------------------------------------------------
void TriMesh::deinitialize()
{
  if (data.pointlist != (double *) NULL) {
    free(data.pointlist);
  }
  if (data.pointattributelist != (double *) NULL) {
    free(data.pointattributelist);
  }
  if (data.pointmarkerlist != (int *) NULL) {
    free(data.pointmarkerlist);
  }
  if (data.segmentlist != (int *) NULL) {
    free(data.segmentlist);
  }
  if (data.segmentmarkerlist != (int *) NULL) {
    free(data.segmentmarkerlist);
  }
  if (data.holelist != (double *) NULL) {
    free(data.holelist);
  }
  if (data.regionlist != (double *) NULL) {
    free(data.regionlist);
  }
  if (data.trianglelist != (int *) NULL){
    free(data.trianglelist);
  }
  if (data.triangleattributelist != (double *) NULL) {
    free(data.triangleattributelist);
  }
  if (data.trianglearealist != (double *) NULL) {
    free(data.trianglearealist);
  }
  if (data.neighborlist != (int *) NULL){
    free(data.neighborlist);
  }
  if (data.edgelist != (int *) NULL) {
    free(data.edgelist);
  }
  if (data.edgemarkerlist != (int *) NULL) {
    free(data.edgemarkerlist);
  }
  if (data.normlist != (double *) NULL) {
    free(data.normlist);
  }
}
//-----------------------------------------------------------------------------
const TriMesh& TriMesh::operator=(const TriMesh& trimesh)
{
  // FIXME: THIS IS DANGEROUS!
  data = trimesh.data;
  return *this;
}
//-----------------------------------------------------------------------------
void TriMesh::print_triangle_parameters()
{
  std::cout << "  Triangle parameters: " << std::endl;
  std::cout << "+----------------------------------------------" << std::endl;     
  std::cout << "| Minmum angle                    (q) : " << minimum_angle << std::endl;
  std::cout << "| Uniform area constraint      (aX.X) : " << uniform_area_constraint << std::endl;
  std::cout << "| Use user defined area function  (u) : " << (area_func !=0 ? "True" : "False" ) << std::endl;
  std::cout << "| Verbosity                       (V) : " << verbosity << std::endl;
  std::cout << "+----------------------------------------------" << std::endl << std::endl;     
}
//-----------------------------------------------------------------------------
// FIXME: Remove when proved not needed
void TriMesh::refine_mesh(TriMesh &output, char* switches)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: TriMesh empty. Nothing to build." );
  output.clear();
  
  triangulate(switches, &data, &output.data, (struct triangulateio *) NULL);
}
//-----------------------------------------------------------------------------
void TriMesh::refine_mesh(TriMesh &output)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: TriMesh empty. Nothing to build." );
  output.clear();
  
  /// Build the switch string
  std::string switches("rze");

  if (minimum_angle > 1.0) 
    {
      switches += "q";
      std::ostringstream o;
      if (!(o << minimum_angle))
	throw std::runtime_error ("Couldn't convert double to string");
      switches += o.str();
    }
  
  if (uniform_area_constraint > 0.0)
    {
      switches += "a";
      std::ostringstream o;
      if (!(o << uniform_area_constraint))
	throw std::runtime_error ("Couldn't convert double to string");
      switches += o.str();
    }
  
  if (area_func != 0)
    {
      switches += "u";
      // Set the global area_function pointer to the one registered in the trimesh
      global_area_function = area_func;
    }

  if ( 0 <= verbosity && verbosity <= 3)
    {
      if (!(verbosity == 1))
	{
	  if (verbosity == 0) 
	    {
	      switches += "Q";
	    }
	  else 
	    for (int i = 1; i <= verbosity; i++ )
	      switches += "V";
	}
    }
  
  // Convert the const char* to a char* 
  char cstr_switches[switches.size()];
  std::strcpy(cstr_switches,switches.c_str());
  
  triangulate(cstr_switches, &data, &output.data, (struct triangulateio *) NULL);
  global_area_function = 0;
}
//-----------------------------------------------------------------------------
void TriMesh::save_triangle(char* file_name)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  FILE *fout;
  char outfilename[FILENAMESIZE];
  int i, j;

  // Save the nodes with any attributes or markers
  sprintf(outfilename, "%s.node", file_name);
  printf("Saving nodes to %s\n", outfilename);
  fout = fopen(outfilename, "w");
  fprintf(fout, "%d  %d  %d  %d\n", data.numberofpoints, 2,
          data.numberofpointattributes, data.pointmarkerlist != NULL ? 1 : 0);
  for (i = 0; i < data.numberofpoints; i++) 
    {
      fprintf(fout, "%d  %.16g  %.16g", i , data.pointlist[i * 2],
              data.pointlist[i * 2 + 1]);
      for (j = 0; j < data.numberofpointattributes; j++) 
	{
	  fprintf(fout, "  %.16g", 
		  data.pointattributelist[i * data.numberofpointattributes + j]);
	}
      if (data.pointmarkerlist != NULL) 
	{
	  fprintf(fout, "  %d", data.pointmarkerlist[i]);
	}
      fprintf(fout, "\n");
    }
  fclose(fout);

  // Save the segments corresponding to the input poly file
  if (data.segmentlist != (int *) NULL){
    sprintf(outfilename, "%s.poly", file_name);
    printf("Saving segments to %s\n", outfilename);
    fout = fopen(outfilename, "w");
    fprintf(fout, "%d  %d  %d  %d\n", 0, 2, data.numberofpointattributes,
	    data.pointmarkerlist != NULL ? 1 : 0);
    fprintf(fout, "%ld  %d\n", static_cast<long int>(data.numberofsegments), 
	    data.segmentmarkerlist != NULL ? 1 : 0);
    for (i = 0; i < data.numberofsegments; i++) 
      {
	fprintf(fout, "%d  %d  %d", i , data.segmentlist[i * 2],
		data.segmentlist[i * 2 + 1]);
	if (data.segmentmarkerlist != NULL) 
	  {
	    fprintf(fout, "  %d", data.segmentmarkerlist[i]);
	  }
	fprintf(fout, "\n");
      }
    
    // FIXME: Cannot print holes and regions due to the hack in the generate funtion in pslg.cpp
    
    // Save holes.
    fprintf(fout, "0\n");
    //fprintf(fout, "%d\n", data.numberofholes);
    //for (i = 0; i < data.numberofholes; i++) {
    //  // Output x, y coordinates.
    //  fprintf(fout, "%d  %.12g  %.12g", i, data.holelist[i * 2],
    //          data.holelist[i * 2 + 1]);
    //  fprintf(fout, "\n");
    //}
    //
    //// Save regions.
    fprintf(fout, "0\n");
    //fprintf(fout, "%d\n", data.numberofregions);
    //for (i = 0; i < data.numberofregions; i++) {
    //  printf("Regions: %d\n",i);
    //  // Output the index, x, y coordinates, attribute (region number)
    //  //   and maximum area constraint (maybe -1).
    //  fprintf(fout, "%d  %.12g  %.12g  %.12g  %.12g\n", i,
    //	    data.regionlist[i * 4], data.regionlist[i * 4 + 1],
    //	    data.regionlist[i * 4 + 2], data.regionlist[i * 4 + 3]);
    //}
    fclose(fout);
  }
  

  // Save the triangles together with any attributes
  sprintf(outfilename, "%s.ele", file_name);
  printf("Saving triangles to %s\n", outfilename);
  fout = fopen(outfilename, "w");
  fprintf(fout, "%d  %d  %d\n", data.numberoftriangles, data.numberofcorners,
          data.numberoftriangleattributes);
  for (i = 0; i < data.numberoftriangles; i++) {
    fprintf(fout, "%d", i);
    for (j = 0; j < data.numberofcorners; j++) {
      fprintf(fout, "  %5d", data.trianglelist[i * data.numberofcorners + j]);
    }
    for (j = 0; j < data.numberoftriangleattributes; j++) {
      fprintf(fout, "  %g",
        data.triangleattributelist[i * data.numberoftriangleattributes + j]);
    }
    fprintf(fout, "\n");
  }

  fclose(fout);
  
}
//-----------------------------------------------------------------------------
void TriMesh::save_dolfin(std::string file_name)
{
  if (!(data.numberofpoints > 0))
    throw std::runtime_error ( "*** Error: Mesh empty. Nothing to save." );
  dolfin::MeshEditor editor;
  dolfin::Mesh dolfin_mesh;
  dolfin::File mesh_file(file_name + ".xml");
      
  std::cout << "Saving mesh to " << file_name << ".xml" << std::endl;

  editor.open(dolfin_mesh,"triangle",2,2);
  editor.init_vertices(data.numberofpoints);
  editor.init_cells(data.numberoftriangles);
  
  // Add vertex coordinates to the dolfin mesh
  for (int i = 0; i < data.numberofpoints; i++)
    {
      editor.add_vertex(i,data.pointlist[i*2],data.pointlist[i*2+1]);
    }
      
  // Add cell information to the dolfin mesh 
  for (int i = 0; i < data.numberoftriangles; i++)
    {
      editor.add_cell(i, data.trianglelist[i*3], data.trianglelist[i*3+1],
		     data.trianglelist[i*3+2]);
    }
  editor.close();
      
  mesh_file << dolfin_mesh;

  // Any point marker to save?
  if (!(data.pointmarkerlist == 0))
    {
      std::cout << "Saving vertex markers to " << file_name
		<< "_vertex_markers.xml" << std::endl;
      dolfin::MeshFunction<unsigned int> vertex_markers(dolfin_mesh,0);
      for (int i = 0; i < data.numberofpoints; i++)
	{
	  vertex_markers[i] = data.pointmarkerlist[i];
	  
	}
      dolfin::File vertex_marker_file(file_name + "_vertex_markers.xml");
      vertex_marker_file << vertex_markers;
    }

  // Any point attributes to save?
  if (!(data.pointattributelist == 0))
    {
      for (int j = 0; j < data.numberofpointattributes; j++)
	{
	  std::cout << "Saving vertex attributes to " << file_name
		    << "_vertex_attributes" << j << ".xml" << std::endl;
	  
	  dolfin::MeshFunction<double> vertex_attributes(dolfin_mesh,0);
	  for (int i = 0; i < data.numberofpoints; i++)
	    {
	      vertex_attributes[i] = data.pointattributelist[i*data.numberofpointattributes+j];
	    }
	  std::ostringstream o;
	  o << j;
	  dolfin::File vertex_attributes_file(file_name + "_vertex_attributes" + o.str() + ".xml");
	  vertex_attributes_file << vertex_attributes;
	}
    }

  // Any face markers to save?
  if (!(data.segmentmarkerlist == 0))
    {
      std::cout << "Saving face markers to " << file_name
		<< "_segment_markers.xml" << std::endl;
      dolfin_mesh.init();
      dolfin::MeshTopology& top = dolfin_mesh.topology();
      dolfin::MeshConnectivity& con01 = top(0,1);
      
      dolfin::MeshFunction<unsigned int> segment_markers(dolfin_mesh,1);

      // Initialize the MeshFunction
      segment_markers = 0;
      
      for (int i = 0; i < data.numberofsegments; i++)
	{
	  if (data.segmentmarkerlist[i]!=0)
	    {
	      // The vertex numbers in the i:th segment
	      int vertex0 = data.segmentlist[i*2];
	      int vertex1 = data.segmentlist[i*2+1];
	      
	      // The segment numbers for each vertex
	      std::vector<unsigned int> segs0(con01(vertex0),con01(vertex0)+con01.size(vertex0));
	      std::vector<unsigned int> segs1(con01(vertex1),con01(vertex1)+con01.size(vertex1));
	      
	      // Sort the triangle indices
	      std::sort(segs0.begin(),segs0.end());
	      std::sort(segs1.begin(),segs1.end());

	      // Find the triangle index that is common for all three vertices
	      std::vector<unsigned int> common_segs(segs0.size());
	      std::vector<unsigned int>::iterator it;
	      it = std::set_intersection (segs0.begin(),segs0.end(),
					  segs1.begin(),segs1.end(),
					  common_segs.begin());
	      
	      if (int(it - common_segs.begin()) == 0)
		{
		  throw std::runtime_error ( "*** Error: No common segments." );
		}
	      if (int(it - common_segs.begin()) > 1)
		{
		  throw std::runtime_error ( "*** Error: To many common segments." );
		}
	      
	      segment_markers[common_segs[0]] = data.segmentmarkerlist[i];
	    }
	}
      dolfin::File segment_marker_file(file_name + "_segment_markers.xml");
      segment_marker_file << segment_markers;
    }
  
  // Any triangle attributes to save?
  if (!(data.numberoftriangleattributes == 0))
    {
      std::cout << "Saving triangle attributes to " << file_name
		<< "_triangle_attributes.xml" << std::endl;
      dolfin::MeshFunction<double> triangle_attributes(dolfin_mesh,2);
      for (int i = 0; i < data.numberoftriangles; i++)
	{
	  triangle_attributes[i] = data.triangleattributelist[i];
	}
      dolfin::File triangle_attribute_file(file_name + "_triangle_attributes.xml");
      triangle_attribute_file << triangle_attributes;
    }
}

